##################################################################
######                                                      ######
######   Icex: automatic extraction of cranial cavities     ######
######                                                      ######
##################################################################


#load libraries
library(Arothron)
library(alphashape3d)
library(Morpho)
library(rgl)
library(Rvcg)

#Load cranial mesh and landmarks
load("IcexData.rda")
sur<-IcexData$sur
set<-IcexData$set

#Load function
Icex<-function(sur,set,planes,keep,param1,param2,alpha,vol=TRUE,plot=TRUE,
               seed=set.seed(123),col1=2,col2=3,center,mode=1,multiple=FALSE){

if(mode==1){
  
  Mesh=sur
  for(i in 1:length(planes)){
    v1<-set[planes[[i]][1],]
    v2<-set[planes[[i]][2],]
    v3<-set[planes[[i]][3],]
    Mesh<-cutMeshPlane(Mesh,v1,v2,v3,keep.upper = keep[i])
    #open3d()
    #wire3d(Mesh,col=i+1)
  }
  
 # wire3d(Mesh,col=2)
  
 IC<-endomaker(Mesh,param1_endo=param1,volume = FALSE,
                 alpha_vol=alpha,param1_ast=param2,plot=FALSE)
 Mesh2<-vcgIsolated(IC$endocast)
 
  }
if(mode==2){
  L<-t(colMeans(set[center,]))
  ast3d_LO<-ext.int.mesh(mesh=sur,param1=param1,default=FALSE, 
                         import_pov = TRUE, 
                         matrix_pov=L,method="ast3d")
  IC<-out.inn.mesh(ast3d_LO,sur,plot=FALSE)
  names(IC)[1]<-"endocast"
  IC$endocast<-vcgIsolated( IC$endocast)
  Mesh2<-vcgIsolated(IC$endocast)
  }
if(mode==3){
    Mesh<-sur
    for(i in 1:length(planes)){
      v1<-set[planes[[i]][1],]
      v2<-set[planes[[i]][2],]
      v3<-set[planes[[i]][3],]
      Mesh<-cutMeshPlane(Mesh,v1,v2,v3,keep.upper = keep[i])
    }
    # wire3d(Mesh,col=2)
    L<-t(colMeans(set[center,]))
    
    ast3d_LO<-ext.int.mesh(mesh=Mesh,param1=param1,default=FALSE, 
                           import_pov = TRUE, 
                           matrix_pov=L,method="ast3d")
    IC<-out.inn.mesh(ast3d_LO,Mesh,plot=FALSE)
    names(IC)[1]<-"endocast"
    IC$endocast<-vcgIsolated( IC$endocast)
    Mesh2<-vcgIsolated(IC$endocast)
  }
if(mode==4){
    Mesh<-sur
    for(i in 1:length(planes)){
      v1<-set[planes[[i]][1],]
      v2<-set[planes[[i]][2],]
      v3<-set[planes[[i]][3],]
      Mesh<-cutMeshPlane(Mesh,v1,v2,v3,keep.upper = keep[i])
    }
    # wire3d(Mesh,col=2)
    ca_lse_<-ext.int.mesh(mesh=Mesh, views=50, param1=param1, default=TRUE, 
                          import_pov = NULL,expand=1, 
                          scale.factor=1,num.cores = NULL)
    IC<-out.inn.mesh(ca_lse_,Mesh,plot=FALSE)
    names(IC)[2]<-"endocast"
    if(multiple==FALSE){
    IC$endocast<-vcgIsolated(IC$endocast)
    Mesh2<-vcgIsolated(IC$endocast)
    }else{
    Mesh2<-IC$endocast
    }
  }

if(vol==TRUE){
conv_endo <- ashape3d(vert2points(Mesh2), alpha =alpha, 
                      pert = TRUE, eps = 1e-06)
x <- conv_endo
selrows = x$triang[, 8 + 1] %in% 2L
tr <- x$triang[selrows, c("tr1", "tr2", "tr3")]
m = (rgl::tmesh3d(vertices = t(x$x), indices = t(tr), homogeneous = FALSE))
sn = alphashape3d::surfaceNormals(x, indexAlpha = 1)
fn = facenormals(m)
dp = dot((fn$normals), t(sn$normals))
m$it[, dp < 0] = m$it[c(1, 3, 2), dp < 0]
conv_sur<- vcgIsolated(m)
bbox <- meshcube(Mesh2)
bbox_w <- sqrt(sum((bbox[1, ] - bbox[3, ])^2))
bbox_h <- sqrt(sum((bbox[1, ] - bbox[2, ])^2))
bbox_l <- sqrt(sum((bbox[1, ] - bbox[5, ])^2))
bbox_c <- prod(bbox_w, bbox_h, bbox_l)
voxelsize <- (bbox_c/1e+05)^(1/3)
xbox <- seq(min(bbox[, 1]), max(bbox[, 1]), by = voxelsize)
ybox <- seq(min(bbox[, 2]), max(bbox[, 2]), by = voxelsize)
zbox <- seq(min(bbox[, 3]), max(bbox[, 3]), by = voxelsize)
GRID <- as.matrix(expand.grid(xbox, ybox, zbox))
DISCR <- vcgClostKD(GRID, Mesh2, sign = TRUE)
voxels_inside_1 <- GRID[which(DISCR$quality <= 0), ]
out_endo <- vcgClostKD(voxels_inside_1, conv_sur, sign = TRUE)
voxels_inside_2 <- voxels_inside_1[which(out_endo$quality >= 
                                           0), ]
count_cells <- nrow(voxels_inside_2)
volume <- ((voxelsize^3) * count_cells)/1000
}
if(vol==TRUE){
out<-list("IC"=IC$endocast,"AlphaShape"=conv_sur,"volume"=volume)}else{
out<-list("IC"=IC$endocast,"AlphaShape"=NULL,"volume"=NULL)  
}
if(plot==TRUE){
layout3d(t(c(1:3)),sharedMouse = TRUE)
shade3d(conv_sur,col=col1,alpha=0.5)
wire3d(IC$endocast,col=col2,add=TRUE)
next3d()
wire3d(IC$endocast,col=col2)
next3d()
shade3d(conv_sur,col=col1,alpha=0.5,specular="black")
}

cat(paste("Save IC?", "(y/n)?"), "\n")
ans <- readLines(n = 1)
if (tolower(ans) == "y") {
mesh2ply(IC$endocast)
if(vol==TRUE){
mesh2ply(conv_sur)}
}
if (tolower(ans) == "n") {
}
return(out)
}
Icv<-function(sur,alpha,col1=2,col2=3,plot=TRUE,vol=TRUE){
    conv_endo <- ashape3d(vert2points(sur), alpha =alpha, 
                          pert = TRUE, eps = 1e-06)
    x <- conv_endo
    selrows = x$triang[, 8 + 1] %in% 2L
    tr <- x$triang[selrows, c("tr1", "tr2", "tr3")]
    m = (rgl::tmesh3d(vertices = t(x$x), indices = t(tr), homogeneous = FALSE))
    sn = alphashape3d::surfaceNormals(x, indexAlpha = 1)
    fn = facenormals(m)
    dp = dot((fn$normals), t(sn$normals))
    m$it[, dp < 0] = m$it[c(1, 3, 2), dp < 0]
    conv_sur<- vcgIsolated(m)
    bbox <- meshcube(sur)
    bbox_w <- sqrt(sum((bbox[1, ] - bbox[3, ])^2))
    bbox_h <- sqrt(sum((bbox[1, ] - bbox[2, ])^2))
    bbox_l <- sqrt(sum((bbox[1, ] - bbox[5, ])^2))
    bbox_c <- prod(bbox_w, bbox_h, bbox_l)
    voxelsize <- (bbox_c/1e+05)^(1/3)
    xbox <- seq(min(bbox[, 1]), max(bbox[, 1]), by = voxelsize)
    ybox <- seq(min(bbox[, 2]), max(bbox[, 2]), by = voxelsize)
    zbox <- seq(min(bbox[, 3]), max(bbox[, 3]), by = voxelsize)
    GRID <- as.matrix(expand.grid(xbox, ybox, zbox))
    DISCR <- vcgClostKD(GRID, sur, sign = TRUE)
    voxels_inside_1 <- GRID[which(DISCR$quality <= 0), ]
    out_endo <- vcgClostKD(voxels_inside_1, conv_sur, sign = TRUE)
    voxels_inside_2 <- voxels_inside_1[which(out_endo$quality >= 
                                               0), ]
    count_cells <- nrow(voxels_inside_2)
    volume <- ((voxelsize^3) * count_cells)/1000
  
  if(vol==TRUE){
    out<-list("IC"=sur,"AlphaShape"=conv_sur,"volume"=volume)}else{
      out<-list("IC"=sur,"AlphaShape"=NULL,"volume"=NULL)
    }
if(plot==TRUE){
  layout3d(t(c(1:3)),sharedMouse = TRUE)
  shade3d(conv_sur,col=col1,alpha=0.5)
  wire3d(sur,col=col2,add=TRUE)
  next3d()
  wire3d(sur,col=col2)
  next3d()
  shade3d(conv_sur,col=col1,alpha=0.5,specular="black")
}
return(out)
}

#Extraction Nasal Cavity
NC<-Icex(sur,set=set,planes=list(c(1,8,9),c(3,10,11),c(1,4,12),c(1,5,13),c(2,8,9)),
     keep=c(TRUE,FALSE,TRUE,FALSE,TRUE),param1=1,param2=4,alpha=20,mode=1)

#Extraction Maxillary sinuses
MS_l<-Icex(sur,set=set,planes=list(c(4,5,8),c(1,8,9),c(1,6,12),c(1,7,13),
 c(4,9,11),c(1,5,11)),keep=c(FALSE,TRUE,TRUE,FALSE,FALSE,TRUE),
 param1=0.1,param2=0.4,alpha=100,mode=1)

MS_r<-Icex(sur,set=set,planes=list(c(4,5,8),c(1,8,9),c(1,6,12),c(1,7,13),
 c(5,8,10),c(1,4,10)),keep=c(FALSE,TRUE,TRUE,FALSE,TRUE,FALSE),
 param1=0.1,param2=0.4,alpha=100,mode=1)

#Extraction Orbits 
OR_l<-Icex(sur,set=set,planes=NULL,keep=NULL,
           param1 = 0.2,param2=NULL,alpha=20,center=c(5,7,15),mode=2)

OR_r<-Icex(sur,set=set,planes=NULL,keep=NULL,
           param1 = 0.2,param2=NULL,alpha=20,center=c(4,6,14),mode=2)

#Extraction Upper Oral Cavity (or "Palate") 
PA<-Icex(sur,set=set,planes=list(c(16,17,18)),keep=FALSE,
           param1=0.5,param2=NULL,alpha=50,center=c(16,17,18),mode=3)


#Extraction Frontal sinuses
FS<-Icex(sur,set=set,planes=list(c(4,5,8),c(3,6,7)),keep=c(TRUE,FALSE),
          param1=2,param2=NULL,alpha=20,center=NULL,mode=4)


